package seu.agent.Firebrigade;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.List;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
import seu.agent.BasicTask.BasicTask;
import seu.agent.BasicTask.ExplorationTask;
import seu.agent.BasicTask.ExtinguishDangerousBuildingTask;
import seu.agent.BasicTask.ExtinguishFireTask;
import seu.agent.BasicTask.ExtinguishFromRefugeTask;
import seu.agent.BasicTask.IsStuckOrBuriedCallHelpTask;
import seu.agent.BasicTask.PartitionExplorationTask;
import seu.agent.BasicTask.RandomWalkTask;
import seu.agent.BasicTask.SaveMyselfTask;
import seu.agent.BasicTask.State;
import seu.agent.BasicTask.WaterRefillTask;
import seu.message.MessagePriority;
import seu.message.Type.BuildingIsBurningMessage;
import seu.message.Type.BuildingIsExploredMessage;
import seu.message.Type.ExtinguishFireTaskMessage;
import seu.message.Type.Message;
import seu.message.Type.MessageCount;
import seu.message.Type.MessageType;
import seu.object.Building.BurningBuildingComparator;
import seu.utilities.DistanceComparator;

/**
 * SEU's Fire Brigade
 */
public class SEUFireBrigade extends SEUAgent<FireBrigade> {

	private static final String MAX_WATER_KEY = "fire.tank.maximum";
	private static final String MAX_DISTANCE_KEY = "fire.extinguish.max-distance";
	private static final String MAX_POWER_KEY = "fire.extinguish.max-sum";

	private int maxDistance;
	private int maxPower;
	private static final int MIN_TOTAL_AREA_TO_CALL_HELP = 150;

	// private ExtinguishFireTask extinguishFireTask;
	public SEUFireBrigade() {
		super();
	}

	/**
	 * Returns the agents number as string.
	 * 
	 * @return Agents number
	 */
	@Override
	public String toString() {
		return "SEU Fire Brigade " + getNo();
	}

	/**
	 * Called after agent is connected to kernel
	 */
	@Override
	protected void postConnect() {
		super.postConnect();
		world.indexClass(StandardEntityURN.BUILDING, StandardEntityURN.REFUGE);
		maxDistance = config.getIntValue(MAX_DISTANCE_KEY);
		maxPower = config.getIntValue(MAX_POWER_KEY);
	}

	@Override
	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.FIRE_BRIGADE);
	}

	/**
	 * Searches the burning buildings from the world model
	 * 
	 * @param sort
	 *            If this boolean is set to true, returning list will contain
	 *            the burning buildings according to distance.
	 * @return The list of burning buildings in the world model
	 */
	public List<Building> getBurningBuildingsSorted() {
		return getBurningBuildings(true, false);
	}

	public List<Building> getBurningBuildings() {
		return getBurningBuildings(false, false);
	}

	public List<Building> getBurningBuildings(boolean noInferno) {
		return getBurningBuildings(false, noInferno);
	}

	public List<Building> getBurningBuildings(boolean sort, boolean noInferno) {
		Collection<Building> e = world.getEntitiesOfType(Building.class,
				StandardEntityURN.BUILDING);
		List<Building> result = new ArrayList<Building>();
		for (Building b : e) {
			if (b.isOnFire()) {
				if (isInMyPartition(b) || isInNeighborPartition(b)) {
					result.add(b);
				} else {
					if (doesNeedHelp(b)) {
						result.add(b);
					}
				}
			}
		}
		if (noInferno) {
			result = removeInfono(result);
		}
		// Sort by distance
		if (sort) {
			Collections.sort(result, new DistanceComparator(location(), world));
		}
		return result;
	}

	protected void processMessage(MessageCount message) {
		int counter = message.getCounter();
		MessageCount reply = new MessageCount(counter + 1);
		sendMessage(reply, MessagePriority.High);
	}

	@Override
	protected void thinkAndAct() {
	}

	protected boolean goToFirePartition() {
		if (state == State.GoingToFirePartition) {
			return false;
		}
		return false;
	}

	public Building nextBuildingToExtinguish() {
		List<Building> burningBuildings = getBurningBuildingsSorted();
		if (burningBuildings.isEmpty()) {
			return null;
		} else {
			return burningBuildings.get(0);
		}
	}

	public Building nextNearBuildingToExtinguish() {
		List<Building> burningSeenBuildings, extinguishableBuildings;
		BurningBuildingComparator comperator;
		Building building;

		burningSeenBuildings = getBurningBuildingsFromChangeSet();
		extinguishableBuildings = new ArrayList<Building>();
		for (Building burningBuilding : burningSeenBuildings) {
			if (canExtinguish(burningBuilding)) {
				extinguishableBuildings.add(burningBuilding);
			}
		}

		if (extinguishableBuildings.isEmpty()) {
			return null;
		} else {
			comperator = new BurningBuildingComparator(world, getTimeStep());
			Collections.sort(extinguishableBuildings, comperator);
			building = extinguishableBuildings.get(0);
			return building;
		}
	}

	@Override
	protected void processMessage(Message message) {
		if (message instanceof BuildingIsExploredMessage) {
			BuildingIsExploredMessage explorationMessage = (BuildingIsExploredMessage) message;
			processExplorationMessage(explorationMessage);
		}

		if (message instanceof BuildingIsBurningMessage) {
			BuildingIsBurningMessage buildingIsBurningMessage = (BuildingIsBurningMessage) message;
			processBuildingIsBurningMessage(buildingIsBurningMessage);
		}

		if (message instanceof ExtinguishFireTaskMessage) {
			ExtinguishFireTaskMessage extinguishFireTaskMessage = (ExtinguishFireTaskMessage) message;
			processExtinguishFireTaskMessage(extinguishFireTaskMessage);
		}

	}

	// 响应中心任务
	@SuppressWarnings("unused")
	protected void processExtinguishFireTaskMessage(
			ExtinguishFireTaskMessage message) {

		EntityID targetBuildingID = new EntityID(message.getBuildingId());
		Building targetBuilding = (Building) world.getEntity(targetBuildingID);

		int agentID = message.getAgentId();
		if (me().getID().getValue() == agentID) {
		}
	}

	// 是否具有救火能力
	// 自己距离房屋的距离大于30米
	// 自己的水量大于0
	public boolean canExtinguish(Building building) {
		int distance;
		boolean canExtinguish;

		if (building == null) {
			return false;
		}
		distance = world.getDistance(me(), building);
		canExtinguish = distance <= maxDistance && canSee(building.getID())
				&& me().getWater() > 0;
		return canExtinguish;
	}

	@Override
	protected boolean canRescue() {
		return false;
	}

	@SuppressWarnings({ "rawtypes" })
	@Override
	public List<BasicTask> getAgentTasks() {

		int maxWater;
		maxWater = config.getIntValue(MAX_WATER_KEY);

		List<BasicTask> tasks;
		tasks = new ArrayList<BasicTask>();

		tasks.add(new IsStuckOrBuriedCallHelpTask(this));
		tasks.add(new ExtinguishFromRefugeTask(this));
		tasks.add(new SaveMyselfTask(this));
		tasks.add(new WaterRefillTask(this, maxWater, 0));
		// tasks.add(new Extinguishtask(this, true));

		tasks.add(new ExtinguishFireTask(this, false));
		tasks.add(new ExtinguishDangerousBuildingTask(this));
		tasks.add(new WaterRefillTask(this, maxWater, maxWater / 2));
		tasks.add(new ExplorationTask(this));
		tasks.add(new PartitionExplorationTask(this));
		tasks.add(new RandomWalkTask(this));
		return tasks;
	}

	// Send Extinguish
	public void sendExtinguish(Building building) {
		sendExtinguish(timeStep, building.getID(), maxPower);
	}

	public boolean doesNeedHelp(Building building) {
		if (building.isFierynessDefined() && building.isTotalAreaDefined()) {

			if ((building.getFierynessEnum() == Fieryness.INFERNO || world
					.getTotalBurningArea(building) > MIN_TOTAL_AREA_TO_CALL_HELP)) {

				// System.out.println("Needs help building " +
				// building.getID().getValue());

				return true;
			}
		}
		return false;

	}

	@Override
	public List<MessageType> getMessagesToListen() {
		List<MessageType> types;

		types = new ArrayList<MessageType>();
		types.add(MessageType.BuildingIsBurningMessage);
		types.add(MessageType.BuildingIsExtinguishedMessage);
		types.add(MessageType.BuildingIsExploredMessage);
		return types;
	}

	public int getMaxDistance() {
		return maxDistance;
	}
}
